"""Implementation of GBA Problem class.

Peter Schubert, HHU Duesseldorf, May 2022
"""

import numpy as np
import re
import types
import sympy
import scipy

from xbanalysis.utils.utils import progress_bar


# Decorator to cache last result
def cache_last(func):
    def wrapper(self, *args):
        fname = wrapper.__name__
        if (fname not in GbaProblem.last_values or
                np.array_equal(GbaProblem.last_values[fname][0], args[0]) is False or
                GbaProblem.cache is False):
            GbaProblem.last_values[fname] = [args[0], func(self, args[0])]
        return GbaProblem.last_values[fname][1]
    wrapper.__name__ = func.__name__
    return wrapper

# TODO: cache last results from rate functions, Jacobians and Hessians
# TODO: call get_reaction_rates from solver ?
# TODO: split get_reaction_rates, get_reaction_jacs, get_reaction_hesses
# TODO: number of mb equations, enzyme transitions density constraints
# number of constraint from len(var_metab_sids), subenz_x_trans.shape[0], 1


class GbaProblem:

    cache = True
    last_values = {}

    def __init__(self, xba_model):
        """Initialize GbaProblem from a xba_model.

        :param xba_model: xba model with all required GBA parameters
        :type xba_model: XbaModel
        """
        self.xba_model = xba_model
        # self.cache = True

        # metabolites and enzymes in the model
        metab_sids = [s.id for s in xba_model.species.values() if s.sboterm == 'SBO:0000247']
        enz_sids = [s.id for s in xba_model.species.values() if s.sboterm in ('SBO:0000250', 'SBO:0000252')]

        # optimization variables
        self.var_metab_sids = [sid for sid in metab_sids if xba_model.species[sid].constant is False]
        self.var_enz_sids = enz_sids
        self.var_sids = self.var_metab_sids + self.var_enz_sids
        self.n_vars = len(self.var_sids)
        self.var_sid2idx = {sid: idx for idx, sid in enumerate(self.var_sids)}
        self.var_mask_metab = [True if sid in self.var_metab_sids else False for sid in self.var_sids]
        self.var_mask_enz = [True if sid in self.var_enz_sids else False for sid in self.var_sids]
        self.var_initial = np.array([xba_model.species[sid].initial_conc for sid in self.var_sids])
        self.var_vols = np.array([xba_model.compartments[xba_model.species[sid].compartment].size
                                  for sid in self.var_sids])

        # constant concentrations
        self.const_metab_sids = [sid for sid in metab_sids if xba_model.species[sid].constant is True]
        self.const_sid2idx = {sid: idx for idx, sid in enumerate(self.const_metab_sids)}

        # protein synthesis reactions get ordered according to the order of their enzyme products
        #  to facilitate matrix multiplication without further remapping
        # - here we expect a one-to-one relationship between enzyme and corresponding protein synthesis reaction
        self.mr_rids = [r.id for r in xba_model.reactions.values()
                        if r.sboterm in ('SBO:0000167', 'SBO:0000179', 'SBO:0000182')]
        self.ps_rids = [xba_model.species[sid].ps_rid for sid in self.var_enz_sids]
        self.mr_vols = np.array([xba_model.compartments[xba_model.reactions[rid].compartment].size
                                 for rid in self.mr_rids])
        self.ps_vols = np.array([xba_model.compartments[xba_model.reactions[rid].compartment].size
                                 for rid in self.ps_rids])

        # calculate stoichiometric sub-matrices and scale them according to volume
        # TODO: sparce matrix storage and calculations
        metab_x_mrs = self.get_stoic_matrix(self.var_metab_sids, self.mr_rids)
        self.metab_x_psrs = self.get_stoic_matrix(self.var_metab_sids, self.ps_rids)
        self.enz_x_mrs = self.get_stoic_matrix(self.var_enz_sids, self.mr_rids)
        self.enz_vols = self.var_vols[self.var_mask_enz]
        metab_vols = self.var_vols[self.var_mask_metab]

        self.metab_x_mras = metab_x_mrs / metab_vols.reshape((-1, 1))
        self.metab_x_psras = self.metab_x_psrs / metab_vols.reshape((-1, 1))
        self.enz_x_mras = self.enz_x_mrs / self.enz_vols.reshape((-1, 1))

        # identify reactions ids for enzyme degradation and corresponding submatrix
        enz_degrad_idxs = self.get_export_idxs(self.enz_x_mrs)
        enz_degrad_rids = np.array(self.mr_rids)[enz_degrad_idxs]
        self.enz_x_degrads = self.enz_x_mrs[:, enz_degrad_idxs]

        # identify reactions ids for enzyme transitions and corresponding submatrix
        # for submatrix perform LU decomposition and remove empty rows from U matrix
        enz_trans_idxs = self.get_trans_idxs(self.enz_x_mras)
        enz_trans_rids = np.array(self.mr_rids)[enz_trans_idxs]
        u = scipy.linalg.lu((self.enz_x_mras[:, enz_trans_idxs]))[2]
        self.subenz_x_trans = u[abs(u).sum(axis=1) > 0]

        # model parameters
        self.model_params = {pid: p.value for pid, p in xba_model.parameters.items()}
        self.model_params.update({cid: c.size for cid, c in xba_model.compartments.items()})
        self.model_params['macro'] = self.model_params.get('macro', 0.0) > 0.0
        self.model_params['ext_conc'] = np.array([xba_model.species[sid].initial_conc
                                                  for sid in self.const_metab_sids])
        self.model_params['np'] = np
        self.model_params['mws'] = np.array([xba_model.species[sid].mw for sid in self.var_sids])

        # create functions
        self.fmras, self.fmras_jac, self.fmras_hesss = self.get_reactions(self.mr_rids, inv=False)
        self.fpstmas, self.fpstmas_jac, self.fpstmas_hesss = self.get_reactions(self.ps_rids, inv=True)
        self.fdegradas, self.fdegradas_jac, self.fdegradas_hesss = self.get_reactions(enz_degrad_rids, inv=False)
        self.fenz_tras, self.fenz_tras_jac, self.fenz_tras_hesss = self.get_reactions(enz_trans_rids, inv=False)
        self.n_constraints = {'heq_mass_balance': len(self.var_metab_sids),
                              'heq_enz_trans': self.subenz_x_trans.shape[0],
                              'heq_density': 1}

    # TODO implement a factory method
    @cache_last
    def mras(self, x):
        return self.fmras(x)

    @cache_last
    def mras_jac(self, x):
        return self.fmras_jac(x)

    @cache_last
    def mras_hesss(self, x):
        return self.fmras_hesss(x)

    @cache_last
    def pstmas(self, x):
        return self.fpstmas(x)

    @cache_last
    def pstmas_jac(self, x):
        return self.fpstmas_jac(x)

    @cache_last
    def pstmas_hesss(self, x):
        return self.fpstmas_hesss(x)

    @cache_last
    def degradas(self, x):
        return self.fdegradas(x)

    @cache_last
    def degradas_jac(self, x):
        return self.fdegradas_jac(x)

    @cache_last
    def degradas_hesss(self, x):
        return self.fdegradas_hesss(x)

    @cache_last
    def enz_tras(self, x):
        return self.fenz_tras(x)

    @cache_last
    def enz_tras_jac(self, x):
        return self.fenz_tras_jac(x)

    @cache_last
    def enz_tras_hesss(self, x):
        return self.fenz_tras_hesss(x)

    def get_stoic_matrix(self, sids, rids):
        """retrieve stoichiometric sub-matrix [sids x rids].

        :param sids: species ids - rows of sub-matrix
        :type sids: list of strings
        :param rids: reaction ids - columns of sub-matrix
        :type rids: list of strings
        :returns: stoic_mat, stoichmetric sub-matrix
        :rtype: np.ndarray
        """
        stoic_mat = np.zeros((len(sids), len(rids)))
        map_metab_row = {sid: idx for idx, sid in enumerate(sids)}
        for col, rid in enumerate(rids):
            for sid, stoic in self.xba_model.reactions[rid].reactants.items():
                if sid in map_metab_row:
                    stoic_mat[map_metab_row[sid], col] = -stoic
            for sid, stoic in self.xba_model.reactions[rid].products.items():
                if sid in map_metab_row:
                    stoic_mat[map_metab_row[sid], col] = stoic
        return stoic_mat

    def update_model_params(self, key, val):
        self.model_params[key] = val

    def _to_vectors(self, rs_str):
        """replace variables and constant metabolites by vector elements.

        Variables converted to x[], external metabolites to ext_conc[].

        :param rs_str: reaction strings (extracted from the xba model).
        :type rs_str: list of strings
        :returns: rs_v_str
        :rtype: list of reaction stings with parameters replaced by vector elements
        """
        rs_v_str = []
        for r_v_str in rs_str:
            for idx, sid in enumerate(self.var_sids):
                r_v_str = re.sub(r'\b' + sid + r'\b', f'x[{idx}]', r_v_str)
            for idx, sid in enumerate(self.const_metab_sids):
                r_v_str = re.sub(r'\b' + sid + r'\b', f'ext_conc[{idx}]', r_v_str)
            rs_v_str.append(r_v_str)
        return rs_v_str

    def _gradients(self, rs_v_str):
        """generate code object for Jacobian and Hessians (lower triangle) of reactions.

        :param rs_v_str: reaction strings with concentrations replaced by vector elements.
        :type rs_v_str: list of string
        :returns: jac
        :rtype: code object to generate Jacobian
        :returns: hess
        :rtype: code object to generate Hessians (lower triangle)
        """
        x = sympy.IndexedBase('x')
        ext_conc = sympy.IndexedBase('ext_conc')
        grads_str = []
        hesss_str = []
        for i, r_v_str in enumerate(rs_v_str):
            func_sym = sympy.sympify(r_v_str, locals={'x': x, 'ext_conc': ext_conc})
            partials_x = []
            hess = []
            for idx1 in range(self.n_vars):
                partial_xsp = func_sym.diff(x[idx1])
                partials_x.append(str(partial_xsp))
                partials_xx = []
                for idx2 in range(self.n_vars):
                    if idx2 <= idx1:
                        partials_xx.append(str(partial_xsp.diff(x[idx2])))
                    else:
                        partials_xx.append('0')
                hess.append('[' + ', '.join(partials_xx) + ']')
            grads_str.append('[' + ', '.join(partials_x) + ']')
            hesss_str.append('[' + ', '.join(hess) + ']')
            progress_bar(i, len(rs_v_str))
        jac = compile('def jac(x): return np.array([' + ', '.join(grads_str) + '])',
                      '<string>', 'exec')
        hesss = compile('def hess(x): return np.array([' + ', '.join(hesss_str) + '])',
                        '<string>', 'exec')
        return jac, hesss

    def get_reactions(self, rids, inv=False):
        """create function objects for reactions rates/times, Jacobian and Hessian

        extracted from kinetic laws for SBML model

        Note: reactions in the SBML model are in amounts per time (not concentration per time)
        I.e. for mass balances equations, we have to devide by volumes where metabolite is located

        :param rids: reaction ids for reactions in amount per time to be processed.
        :type rids: list of strings
        :param inv: flag indicating if inverse of reaction required (i.e. reaction times)
        :type inv: boolean (default: False)
        :returns: ras
        :rtype: function object, returning an 1D array of function results
        :returns: ras_jac
        :rtype: function object, returning a 2D array with the Jacobian matrix ∂/∂xi
        :returns: ras_hesss
        :rtype: function object, returning a 3D array (or sparce/lower triangele) of the Hessian
                matrices for each individual function
        """
        if inv is False:
            kinetics_str = [self.xba_model.reactions[rid].expanded_kl for rid in rids]
        else:
            kinetics_str = [f'1.0/({self.xba_model.reactions[rid].expanded_kl})' for rid in rids]
        v_str = self._to_vectors(kinetics_str)
        func_code = compile('def reactions(x): return np.array([' + ', '.join(v_str) + '])',
                            '<string>', 'exec')
        jac_code, hesss_code = self._gradients(v_str)
        ras = types.FunctionType(func_code.co_consts[0], self.model_params)
        ras_jac = types.FunctionType(jac_code.co_consts[0], self.model_params)
        ras_hesss = types.FunctionType(hesss_code.co_consts[0], self.model_params)
        return ras, ras_jac, ras_hesss

    @staticmethod
    def get_trans_idxs(smat):
        """Identify metabolic reactions in S matrix with species transitions.

        I.e. columns containing both positive and negative entries

        :param smat:
        :type smat:
        :return: bolean vector identifying relevant columns
        :rtype: 1D ndarray with dtype bool
        """
        mask_trans_idxs = np.zeros(smat.shape[1])
        for col_idx in range(smat.shape[1]):
            if ((np.count_nonzero(smat[:, col_idx] < 0) > 0) and
                    np.count_nonzero(smat[:, col_idx] > 0) > 0):
                mask_trans_idxs[col_idx] = 1.0
        return mask_trans_idxs > 0

    @staticmethod
    def get_export_idxs(smat):
        """Identify metabolic reactions in S matrix with consumption only.

        I.e. columns containing only negative entries

        :param smat:
        :type smat:
        :return: bolean vector identifying relevant columns
        :rtype: 1D ndarray with dtype bool
        """
        mask_remove_idxs = np.zeros(smat.shape[1])
        for col_idx in range(smat.shape[1]):
            if ((np.count_nonzero(smat[:, col_idx] < 0) > 0) and
                    np.count_nonzero(smat[:, col_idx] > 0) == 0):
                mask_remove_idxs[col_idx] = 1.0
        return mask_remove_idxs > 0

    @cache_last
    def get_growth_rate_0(self, x):
        """Calculate growth rate at x, assuming no protein degradation

        Equations:
            gr0(x) = 1/tau(x)
            tau(x) = pstms(x).dot(ce)
            pstms(x) = enz_vols * ptsmas(x)

        :param x: optimization variables
        :type x: numpy.ndarray
        :returns: growth rate at x (without protein degradation)
        :rtype: float
        """
        ce_var = x[self.var_mask_enz]
        pstms = self.enz_vols * self.pstmas(x)

        gr0 = 1.0 / pstms.dot(ce_var)
        return gr0

    @cache_last
    def get_growth_rate_0_grad(self, x):
        """Calculate growth rate gradient at x, assuming no protein degradation

        Equations:
            gr0(x) = 1/tau(x)
            tau(x) = pstms(x).dot(ce)
            pstms(x) = enz_vols * ptsmas(x)

            gr0_grad(x) = -tau_grad(x)/tau(x)^2  = -gr0(x)^2 * tau_grad(x)
            tau_grad(x) = ce.dot(pstms_jac(x)) + pstms(x).dot(ce_jac)
            pstms_jac(x) = enz_vols.reshape(-1,1) * pstmas(x)

        :param x: optimization variables
        :type x: numpy.ndarray
        :returns: growth rate gradient at x (without protein degradation)
        :rtype: numpy.ndarray (1D)
        """
        gr0 = self.get_growth_rate_0(x)
        ce = x[self.var_mask_enz]
        ce_jac = np.hstack((np.zeros((len(ce), len(x) - len(ce))), np.eye(len(ce))))
        pstms = self.enz_vols * self.pstmas(x)
        pstms_jac = self.enz_vols.reshape((-1, 1)) * self.pstmas_jac(x)

        tau_grad = ce.dot(pstms_jac) + pstms.dot(ce_jac)
        gr0_grad = - gr0**2 * tau_grad
        return np.nan_to_num(gr0_grad, nan=0.0)

    @cache_last
    def get_growth_rate_0_hess(self, x):
        """Calculate growth rate Hessian at x, assuming no protein degradation

        Equations:
            gr0(x) = 1/tau(x)
            tau(x) = pstms(x).dot(ce)
            pstms(x) = enz_vols * ptsmas(x)

            gr0_grad(x) = -tau_grad(x)/tau(x)^2  = -gr0(x)^2 * tau_grad(x)
            tau_grad(x) = ce.dot(pstms_jac(x)) + pstms(x).dot(ce_jac)
            pstms_jac(x) = enz_vols.reshape(-1,1) * pstmas(x)

            gr0_hess(x) = -2*gr0(x) * gr0_grad(x) * tau_grad(x) - gr0(x)^2 * tau_hess(x)
            tau_hess = ce_jac.dot(pstms_jac(x)) + ce.dot(pstms_hesss(x)) + pstms_jac(x).dot(ce_jac)
            ce_hess = 0

        :param x: optimization variables
        :type x: numpy.ndarray
        :returns: growth rate Hessian at x (without protein degradation)
        :rtype: numpy.ndarray (2D) with shape [x,x]
        """
        gr0 = self.get_growth_rate_0(x)
        gr0_grad = self.get_growth_rate_0_grad(x)
        ce = x[self.var_mask_enz]
        ce_jac = np.hstack((np.zeros((len(ce), len(x) - len(ce))), np.eye(len(ce))))
        pstms = self.enz_vols * self.pstmas(x)
        pstms_jac = self.enz_vols.reshape((-1, 1)) * self.pstmas_jac(x)
        pstms_hesss = self.enz_vols.reshape((-1, 1, 1)) * self.pstmas_hesss(x)

        tau_grad = ce.dot(pstms_jac) + pstms.dot(ce_jac)
        hess_a = np.outer(gr0_grad, tau_grad)
        hess_b1 = (ce.reshape((-1, 1, 1))*pstms_hesss).sum(axis=0)
        hess_b2 = ce_jac.T.dot(pstms_jac)
        hess_b3 = pstms_jac.T.dot(ce_jac)
        gr0_hess = - 2 * gr0 * hess_a - gr0**2 * (hess_b1 + hess_b2 + hess_b3)
        return np.nan_to_num(np.tril(gr0_hess), nan=0.0)

    @cache_last
    def get_growth_rate(self, x):
        """Calculate growth rate at x, considering protein degradation

        considering stoichiometric coefficiens unequal unity for enzyme turnover
        stoichiometric sub-matrices do not require volume scaling (as Vol_enz gets factored out)
        With dilution (negative stoichiometric coefficients in enz_x_mrs), growth rate reduces

        Equations:
            gr(x) = gr0(x) * (1 + enz_to_mras(x).dot(pstmas(x)))
            enz_to_mras(x) = enz_x_mrs.dot(mras(x))

        :param x: optimization variables
        :type x: numpy.ndarray
        :returns: growth rate at x (considering protein degradation)
        :rtype: float
        """
        gr0 = self.get_growth_rate_0(x)
        if np.count_nonzero(self.enz_x_mrs) == 0:
            gr = gr0
        else:
            pstmas = self.pstmas(x)
            enz_to_mras = self.enz_x_degrads.dot(self.degradas(x))

            gr = gr0 * (1.0 + enz_to_mras.dot(pstmas))
        return gr

    @cache_last
    def get_growth_rate_grad(self, x):
        """Calculate growth rate gradient at x, considering protein degradation

        Equations:
            gr(x) = gr0(x) * (1 + enz_to_mras(x).dot(pstmas(x)))
            enz_to_mras(x) = enz_x_mrs.dot(mras(x))

            gr_grad(x) = gr0_grad(x) * (1 + enz_to_mras(x).dot(pstmas(x)))
                         + gr0(x) * (pstmas(x).dot(enz_to_mras_jac(x)) + enz_to_mras(x).dot(pstmas_jac(x)))
            enz_to_mras_jac(x) = enz_to_mrs.dot(mras_jac(x))

        :param x: optimization variables
        :type x: numpy.ndarray
        :returns: growth rate gradient at x (considering protein degradation)
        :rtype: numpy.ndarray (1D)
        """
        gr0_grad = self.get_growth_rate_0_grad(x)
        if np.count_nonzero(self.enz_x_mrs) == 0:
            gr_grad = gr0_grad
        else:
            gr0 = self.get_growth_rate_0(x)
            pstmas = self.pstmas(x)
            pstmas_jac = self.pstmas_jac(x)
            enz_to_mras = self.enz_x_degrads.dot(self.degradas(x))
            enz_to_mras_jac = self.enz_x_degrads.dot(self.degradas_jac(x))

            gr_grad = (gr0_grad * (1.0 + enz_to_mras.dot(pstmas))
                       + gr0 * (pstmas.dot(enz_to_mras_jac) + enz_to_mras.dot(pstmas_jac)))
        return np.nan_to_num(gr_grad, nan=0.0)

    @cache_last
    def get_growth_rate_hess(self, x):
        """Calculate growth rate Hessian at x, considering protein degradation

         Equations:
            gr(x) = gr0(x) * (1 + enz_to_mras(x).dot(pstmas(x)))
            enz_to_mras(x) = enz_x_mrs.dot(mras(x))

            gr_grad(x) = gr0_grad(x) * (1 + enz_to_mras(x).dot(pstmas(x)))
                         + gr0(x) * (pstmas(x).dot(enz_to_mras_jac(x)) + enz_to_mras(x).dot(pstmas_jac(x)))
            enz_to_mras_jac(x) = enz_to_mrs.dot(mras_jac(x))

            gr_hess(x) = gr0_hess(x) * (1 + enz_to_mras(x).dot(pstmas(x)))
                         + (pstmas(x).dot(enz_to_mras_jac(x)) + enz_to_mras(x).dot(pstmas_jac(x))) * gr0_grad(x)
                         + gr0_grad(x) * (pstmas(x).dot(enz_to_mras_jac(x)) + enz_to_mras(x).dot(pstmas_jac(x)))
                         + gr0(x) * (pstmas_jac(x).dot(enz_to_mras_jac(x))
                                     + enz_to_mras_hess(x).dot(pstmas(x))
                                     + enz_to_mras_jac(x).dot(pstmas_jac(x))
                                     + pstmas_hess(x).dot(enz_to_mras(x))
                                     )

        :param x: optimization variables
        :type x: numpy.ndarray
        :returns: growth rate Hessian at x (considering protein degradation)
        :rtype: numpy.ndarray (2D) with shape [x,x]
        """
        gr0_hess = self.get_growth_rate_0_hess(x)
        if np.count_nonzero(self.enz_x_mrs) == 0:
            gr_hess = gr0_hess
        else:
            gr0 = self.get_growth_rate_0(x)
            gr0_grad = self.get_growth_rate_0_grad(x)
            pstmas = self.pstmas(x)
            pstmas_jac = self.pstmas_jac(x)
            pstmas_hesss = self.pstmas_hesss(x)
            enz_to_mras = self.enz_x_degrads.dot(self.degradas(x))
            enz_to_mras_jac = self.enz_x_degrads.dot(self.degradas_jac(x))
            enz_to_mras_hesss = np.array([(row.reshape((-1, 1, 1)) * self.degradas_hesss(x)).sum(axis=0)
                                          for row in self.enz_x_degrads])

            hess_a = gr0_hess * (1.0 + enz_to_mras.dot(pstmas))
            grad = pstmas.dot(enz_to_mras_jac) + enz_to_mras.dot(pstmas_jac)
            hess_b1 = np.outer(gr0_grad, grad)
            hess_b2 = np.outer(grad, gr0_grad)
            hess_c1 = pstmas_jac.T.dot(enz_to_mras_jac)
            hess_c2 = (pstmas.reshape((-1, 1, 1)) * enz_to_mras_hesss).sum(axis=0)
            hess_c3 = enz_to_mras_jac.T.dot(pstmas_jac)
            hess_c4 = (enz_to_mras.reshape((-1, 1, 1)) * pstmas_hesss).sum(axis=0)
            gr_hess = hess_a + hess_b1 + hess_b2 + gr0 * (hess_c1 + hess_c2 + hess_c3 + hess_c4)
        return np.nan_to_num(np.tril(gr_hess), nan=0.0)

    @cache_last
    def get_alphas(self, x):
        """Calculate ribosomal allocations at x, considering protein degradation

        alpha_e = gr * ce_e * pstm_e + p_degr_e * pstm_e
        alpha = gr * ce.dot(pstms(x)) + p_degrs(x).dot(pstms(x))

        :param x: optimization variables
        :type x: numpy.ndarray
        :returns: ribosomal allocations at x (considering protein degradation)
        :rtype: numpy.ndarray (1D)
        """
        ce = x[self.var_mask_enz]
        gr = self.get_growth_rate(x)
        pstms = self.enz_vols * self.pstmas(x)

        alphas0 = gr * ce * pstms
        if np.count_nonzero(self.enz_x_mrs) == 0:
            alphas = alphas0
        else:
            enz_to_mras = self.enz_x_degrads.dot(self.degradas(x))
            alphas = alphas0 - enz_to_mras * self.pstmas(x)
        return alphas

    @cache_last
    def heq_mass_balance(self, x):
        """Calculate mass balance equations at x for metabolites

        incuding protein degradation, which consume metabolites
        Note: protein degradation is also part of metabolic reactions (mras), which return
              metabolites.

        metab_mb = dcm/dt = metabolic_balance + growth_dilution + sythesis_enzyme_balance
          metabolic_balance = metab_x_mras.dot(mras(x))
          growth_dilution = λ * (metab_x_psrs.dot(ce) - cm)
            ce_scaled = enz_vols * ce   # convert concentration to amounts, considering stoichiometry
          sythesis_enzyme_balance = - metab_x_psras.dot(enz_vols * enzyme_balance)
            enzyme_balance = enz_x_mras.dot(mras(x))

        :param x: optimization variables
        :type x: numpy.ndarray
        :returns: mass balance calculated at x
        :rtype: numpy.ndarray (1D) with shape (len(var_mask_metab), 1)
        """
        gr = self.get_growth_rate(x)
        cm = x[self.var_mask_metab]
        ce = x[self.var_mask_enz]

        metabolic_mb = self.metab_x_mras.dot(self.mras(x))
        dilution = self.metab_x_psrs.dot(ce) - cm
        metab_mb = metabolic_mb + gr * dilution
        if np.count_nonzero(self.enz_x_mras) > 0:
            enz_mb = self.enz_x_mras.dot(self.mras(x))
            metab_mb -= self.metab_x_psrs.dot(enz_mb)
        return metab_mb

    @cache_last
    def heq_mass_balance_jac(self, x):
        """Calculate Jacobian of mass balance equations at x for metabolites

        incuding protein degradation , which consume metabolites
        Note: protein degradation is also part of metabolic reactions (mras), which return
              metabolites.

        metab_mb = dcm/dt = metabolic_mb + growth_dilution + sythesis_enzyme_balance
          metabolic_mb = metab_x_mras.dot(mras(x))
          growth_dilution = λ * (metab_x_psras.dot*(ce_scaled) - cm)
            ce_scaled = enz_vols * ce   # convert concentration to amounts, considering stoichiometry
          sythesis_enzyme_balance = - metab_x_psras.dot(enz_vols * enzyme_balance)
            enzyme_balance = enz_x_mras.dot(mras(x))

        ∂metab_mb/∂x = metabolic_mb_jac + dilution_jac + sythesis_enzyme_balance_jac
          metabolic_mb_jac = metab_x_mras.dot(mras_jac(x))
          dilution_jac = gr_grad * (metab_x_psras.dot(ce_scaled) - cm)
                         + gr * (metab_x_psras.dot(ce_scaled_jac) - cm_jac)
            ce_scaled = enz_vols * ce   # convert concentration to amounts, considering stoichiometry
            ce_jac = np.hstack( (np.zeros((len(ce), len(cm))), np.eye(len(ce)) ))
            cm_jac = np.hstack( (np.eye(len(cm)), np.zeros((len(cm), len(ce))) ))
            ce_scaled_jac = enz_vols.reshape((-1,1)) * ce_jac
          sythesis_enzyme_balance_jac = - metab_x_psras.dot(enz_vols * enzyme_balance_jac)
            enzyme_balance_jac = enz_x_mras.dot(mras_jac(x))

        :param x: optimization variables
        :type x: numpy.ndarray
        :returns: Jacobian of mass balance equations at x
        :rtype: numpy.ndarray (2D)
        """
        gr = self.get_growth_rate(x)
        gr_grad = self.get_growth_rate_grad(x)
        cm = x[self.var_mask_metab]
        ce = x[self.var_mask_enz]
        cm_jac = np.hstack((np.eye(len(cm)), np.zeros((len(cm), len(x) - len(cm)))))
        ce_jac = np.hstack((np.zeros((len(ce), len(x) - len(ce))), np.eye(len(ce))))
        mras_jac = self.mras_jac(x)

        metabolic_jac = self.metab_x_mras.dot(mras_jac)
        dilution = self.metab_x_psrs.dot(ce) - cm
        dilute_1_jac = np.outer(dilution, gr_grad)
        dilute_2_jac = gr * (self.metab_x_psrs.dot(ce_jac) - cm_jac)

        metab_mb_jac = metabolic_jac + dilute_1_jac + dilute_2_jac
        if np.count_nonzero(self.enz_x_mras) > 0:
            enz_mb_jac = self.enz_x_mras.dot(mras_jac)
            metab_mb_jac -= self.metab_x_psrs.dot(enz_mb_jac)
        return np.nan_to_num(metab_mb_jac, nan=0.0)

    @cache_last
    def heq_mass_balance_hess(self, x):
        """Calculate Hessians of mass balance equations at x

        incuding protein degradation , which consume metabolites
        Here: assuming protein degradation has stoichometry of -1 for relevant enzyme
        Note: protein degradation is also part of metabolic reactions (mras), which return
              metabolites.

        array of n Hessians
        metab_mb_hesss = metabolic_mb_hess + dilution_hess + sythesis_enzyme_balance_hess
          metabolic_mb_hesss = metab_x_mras.dot(mras_hesss(x))
          dilution_hessd = gr_hess * (metab_x_psras.dot(ce_scaled) - cm)
                          + gr_grad * (metab_x_psras.dot(ce_scaled_jac) - cm_jac)
                          + gr_grad * (metab_x_psras.dot(ce_scaled_jac) - cm_jac)
                          + gr *  (metab_x_psras.dot(ce_scaled_hess) - cm_hess)
          dilution_hesss = gr_hess * (metab_x_psras.dot(ce_scaled) - cm)
                          + 2 *  gr_grad * (metab_x_psras.dot(ce_scaled_jac) - cm_jac)

            ce_scaled = enz_vols * ce   # convert concentration to amounts, considering stoichiometry
            ce_jac = np.hstack( (np.zeros((len(ce), len(cm))), np.eye(len(ce)) ))
            cm_jac = np.hstack( (np.eye(len(cm)), np.zeros((len(cm), len(ce))) ))
            ce_scaled_jac = enz_vols.reshape((-1,1)) * ce_jac
            cm_hess = np.zeros((x,x)) # as cm_jac only contains constants
            ce_hess = np.zeros((x,x))

          sythesis_enzyme_balance_hesss = - metab_x_psras.dot(enz_vols * enzyme_balance_hesss)

            enzyme_balance_jac = enz_x_mras.dot(mras_hesss(x))

          sythesis_enzyme_balance_jac = - metab_x_psras.dot(enz_vols * enzyme_balance_jac)
            enzyme_balance_jac = enz_x_mras.dot(mras_jac(x))

        :param x: optimization variables
        :type x: numpy.ndarray
        :returns: Hessians of mass balance equations at x
        :rtype: numpy.ndarray (3D) with shape [n,x,x]
        """
        gr_grad = self.get_growth_rate_grad(x)
        gr_hess = self.get_growth_rate_hess(x)
        cm = x[self.var_mask_metab]
        ce = x[self.var_mask_enz]
        cm_jac = np.hstack((np.eye(len(cm)), np.zeros((len(cm), len(x) - len(cm)))))
        ce_jac = np.hstack((np.zeros((len(ce), len(x) - len(ce))), np.eye(len(ce))))
        mras_hesss = self.mras_hesss(x)
        metabolic_hesss = np.array([(row.reshape((-1, 1, 1)) * mras_hesss).sum(axis=0)
                                    for row in self.metab_x_mras])
        dilution = self.metab_x_psrs.dot(ce) - cm
        dilution_jac = self.metab_x_psrs.dot(ce_jac) - cm_jac

        dilution_1a_hesss = dilution.reshape((-1, 1, 1)) * gr_hess
        dilution_1b_hesss = np.array([np.outer(row, gr_grad) for row in dilution_jac])
        dilution_2a_hesss = np.array([np.outer(gr_grad, row) for row in dilution_jac])
        # dilution_2b_hesss = np.zeros_like(metabolic_hesss)

        metab_mb_hesss = metabolic_hesss + dilution_1a_hesss + dilution_1b_hesss + dilution_2a_hesss
        if np.count_nonzero(self.enz_x_mras) > 0:
            enz_mb_hesss = np.array([(row.reshape((-1, 1, 1)) * mras_hesss).sum(axis=0)
                                     for row in self.enz_x_mras])
            metab_mb_hesss -= np.array([(row.reshape((-1, 1, 1)) * enz_mb_hesss).sum(axis=0)
                                        for row in self.metab_x_psrs])
        return np.nan_to_num(np.tril(metab_mb_hesss), nan=0.0)

    @cache_last
    def heq_enz_trans(self, x):
        """Calculate enzyme transition equations at x

        :param x: optimization variables
        :type x: numpy.ndarray
        :returns: Jacobian of mass balance equations at x
        :rtype: numpy.ndarray (1D) with shape (len(subenz_x_trans.shape[1]), 1)
        """
        return self.subenz_x_trans.dot(self.enz_tras(x))

    @cache_last
    def heq_enz_trans_jac(self, x):
        """Calculate Jacobian for enzyme transition equations at x

        :param x: optimization variables
        :type x: numpy.ndarray
        :returns: Jacobian of mass balance equations at x
        :rtype: numpy.ndarray (2D)
        """
        return self.subenz_x_trans.dot(self.enz_tras_jac(x))

    @cache_last
    def heq_enz_trans_hess(self, x):
        """Calculate Hessians of enzyme transition equations at x

        :param x: optimization variables
        :type x: numpy.ndarray
        :returns: Hessians of mass balance equations at x
        :rtype: numpy.ndarray (3D) with shape [n,x,x]
        """
        return np.array([(row.reshape((-1, 1, 1)) * self.enz_tras_hesss(x)).sum(axis=0)
                         for row in self.subenz_x_trans])

    @cache_last
    def heq_density(self, x):
        """Calculate density (g/l) constraint at x

        Could be extended having several density constraints (not yet implemented)

        macromolecular density is total protein mass density
        alternalively, dry mass density also adds total metabolite mass density

        using ipopt, the constraint bounds could be set to rho (only add up masses)
        density = sum(species_concentration * molecular_weight)

        :param x: optimization variables
        :type x: numpy.ndarray
        :returns: mass balance calculated at x
        :rtype: numpy.ndarray (1D) with shape (1,) for now
        """
        ce = x[self.var_mask_enz]
        if self.model_params.get('macro', True) is True:
            density = self.model_params['mws'][self.var_mask_enz].dot(ce)
        else:
            density = self.model_params['mws'].dot(x)
        return np.array([self.model_params['rho'] - density])

    @cache_last
    def heq_density_jac(self, x):
        """Calculate gradient of density constraint at x

        for single density, this returns a gradient vector
        for multiple densities (not implemented), this
        would return a Jacobian

        :param x: optimization variables
        :type x: numpy.ndarray
        :returns: mass balance jacobian at x
        :rtype: numpy.ndarray (1D)
        """
        if self.model_params.get('macro', True) is True:
            density_grad = np.zeros(len(x))
            density_grad[self.var_mask_enz] = -self.model_params['mws'][self.var_mask_enz]
        else:
            density_grad = -self.model_params['mws']
        return density_grad

    # noinspection PyUnusedLocal
    @cache_last
    def heq_density_hess(self, x):
        """Calculate hessian of density constraint at x

        :param x: optimization variables
        :type x: numpy.ndarray
        :returns: mass balance Hessian at x
        :rtype: numpy.ndarray (3D), shape [1,x,x]
        """
        hess = np.zeros((1, self.n_vars, self.n_vars))
        return np.tril(hess)
